#pragma once
#include <map>
#include <unordered_set>
#include <vector>
#include "Exception.h"
#include "Types.h"
#include "Print_Functions.h"
#include <algorithm>


template<uint8_t DIM>
class MeshLoader
{
protected:
	std::vector<Node<DIM>> nodes;
	std::vector<FiniteElement> fin_els;
	std::multimap<uint32_t, BoundaryFiniteElement> b_fin_els;
	

	//Load mesh from file
	virtual void loadMesh(const std::string& filename) = 0;


	//Adds node to "nodes" container
    //       ?
    //        
	Node<DIM>& AddNode(Node<DIM>& cur_node)
	{
		nodes.push_back(cur_node);
        //     ?      ,
        //      Ѩ
		return cur_node;
	}


	//Adds FE to "fe_els" container
    //    
	FiniteElement& AddFE(FiniteElement& cur_fe)
	{
		fin_els.push_back(cur_fe);
		return cur_fe;
	}

	
	//Adds BFE to "fe_els" container
    //  
	BoundaryFiniteElement& AddBFE(const uint32_t& bound_id, BoundaryFiniteElement& cur_bfe)
	{
		b_fin_els.emplace(bound_id, cur_bfe);
		return cur_bfe;
	}
	

private:

	Node<DIM> createMiddleNode(const Node<DIM>& node1, const Node<DIM>& node2)
	{
		Node<DIM> new_node;

		new_node.id = nodes.size() + 1;

		for (uint8_t i = 0; i < DIM; ++i)
		{
			new_node.coord[i] = (node1.coord[i] + node2.coord[i]) / 2;
		}

		if ((node1.in_node == true) && (node2.in_node == true))
		{
			new_node.in_node = true;
		}
		else
		{
			new_node.in_node = false;
		}

		return new_node;
	}


public:

	MeshLoader() {};

	//Returns nodes container
	const std::vector<Node<DIM>>& getNodeContainer() const noexcept { return nodes; }


	//Returns finite elements container
	const std::vector<FiniteElement>& getFEContainer() const noexcept { return fin_els; }


	//Returns boundary finite elements container
	const std::multimap<uint32_t, BoundaryFiniteElement>& getBFEContainer() const noexcept { return b_fin_els; }

	
	//Searches finite elements with nodes id1, id2, id3
	std::vector<FiniteElement> findFE(const uint32_t& id1, const uint32_t& id2, const uint32_t& id3) const
	{
		std::vector<FiniteElement> buffer;
        //   template?   
		std::vector<FiniteElement>::template const_iterator cur_fe_it = fin_els.begin();
		uint8_t fe_vertex_cnt = fin_els.begin()->nodes.size(); //we take size of finite element vector "nodes"(we take it from first element just for simplicity)


		while (cur_fe_it != fin_els.end())
		{
			cur_fe_it = std::find_if(cur_fe_it, fin_els.end(),
			[id1, id2, id3, fe_vertex_cnt](const FiniteElement& ordered_fe)
			{
				uint8_t cnt = 0; //amount of matched nodes
				for (uint8_t i = 0; i < fe_vertex_cnt; ++i)
				{
					if ((ordered_fe.nodes[i] == id1) || (ordered_fe.nodes[i] == id2) || (ordered_fe.nodes[i] == id3))
					{
						cnt++;
					}
				}
                //       :
                // return cnt == 3;
				if (cnt == 3) //if "ordered_fe" has nodes id1, id2, id3
				{
					return true;
				}
				else
				{
					return false;
				}
			});

			if (cur_fe_it != fin_els.end())
			{
				buffer.push_back(*cur_fe_it);
				cur_fe_it++;
			}
		}

		if (buffer.size() == 0)
			throw EmptyBufferExc();
		else
		{
			//PRINT FOUNDED ELEMENTS
			printFE(buffer.begin(), buffer.end());
		}

		return buffer;
	}


	//Searches finite elements with nodes id1, id2
	std::vector<FiniteElement> findFE(const uint32_t& id1, const uint32_t& id2) const
	{
		std::vector<FiniteElement> buffer;
		std::vector<FiniteElement>::template const_iterator cur_fe_it = fin_els.begin();
		uint8_t fe_vertex_cnt = fin_els.begin()->nodes.size(); //we take size of finite element vector "nodes"(we take it from first element just for simplicity)


		while (cur_fe_it != fin_els.end())
		{
			cur_fe_it = std::find_if(cur_fe_it, fin_els.end(),
				[id1, id2, fe_vertex_cnt](const FiniteElement& ordered_fe)
			{
				uint8_t cnt = 0; //amount of matched nodes
				for (uint8_t i = 0; i < fe_vertex_cnt; ++i)
				{
					if ((ordered_fe.nodes[i] == id1) || (ordered_fe.nodes[i] == id2))
					{
						cnt++;
					}
				}
				if (cnt == 2) //if "ordered_fe" has nodes id1, id2
				{
					return true;
				}
				else
				{
					return false;
				}
			});

			if (cur_fe_it != fin_els.end())
			{
				buffer.push_back(*cur_fe_it);
				cur_fe_it++;
			}
		}

		if (buffer.size() == 0)
			throw EmptyBufferExc();
		else
		{
			//PRINT FOUNDED ELEMENTS
			printFE(buffer.begin(), buffer.end());
		}

		return buffer;
	}


	//Returns container with nodes on boundary "boundary_id"(   ,    "boundary_id")
	std::vector<Node<DIM>> getBoundaryNodes(const uint32_t boundary_id)
	{
        //     ,  
        //  set,         
		std::vector<Node<DIM>> buffer;

		auto range = b_fin_els.equal_range(boundary_id); //"range" is pair of iterators pointing on begin and end of range of elements with key "boundary_id"

		for (auto it = range.first; it != range.second; it++)
		{
			for (uint8_t i = 0; i < it->second.nodes.size(); ++i)
			{
				buffer.push_back(getNodeWithID(it->second.nodes[i]));
			}
		}

		
		auto comp = [](const Node<DIM>& n1, const Node<DIM>& n2)
		{
            // return n1.id < n2.id;
			if (n1.id < n2.id) return true;
			else return false;
		};

		auto equal = [](const Node<DIM>& n1, const Node<DIM>& n2)
		{
            // return n1.id == n2.id;
            if (n1.id == n2.id) return true;
			else return false;
		};

		std::sort(buffer.begin(), buffer.end(), comp);
		auto last = std::unique(buffer.begin(), buffer.end(), equal);
		buffer.erase(last, buffer.end()); //erase incorrect elements
		

		if (buffer.size() == 0)
			throw EmptyBufferExc();
		else
		{
			//PRINT FOUNDED ELEMENTS
			printNode<DIM>(buffer.begin(), buffer.end());
		}

		return buffer;
	}

	
	//Returns FE container with id "mater_id" of material
	std::vector<FiniteElement> getMaterialFEs(const uint16_t&  mater_id) const 
	{
		std::vector<FiniteElement> buffer;


		for (auto it = fin_els.begin(); it != fin_els.end(); ++it)
		{
			if (it->material_id == mater_id)
			{
				buffer.push_back(*it);
			}
		}

		if (buffer.size() == 0)
			throw EmptyBufferExc();
		else
		{
			//PRINT FOUNDED ELEMENTS
			printFE(buffer.begin(), buffer.end());
		}

		return buffer;
	}
	

	//Returns container with BFEs with boundary id "boundary_id"(    ,    "boundary_id")
	std::vector<BoundaryFiniteElement> getBoundaryBFEs(const uint32_t boundary_id) const
	{
		std::vector<BoundaryFiniteElement> buffer;


		auto range = b_fin_els.equal_range(boundary_id);
		for (auto& it = range.first; it != range.second; ++it)
			buffer.push_back(it->second);

		if (buffer.empty())
		{
			throw EmptyBufferExc();
		}
		else
		{
			//PRINT ALL ELEMENTS
			printBFE(buffer.begin(), buffer.end());
		}

		return buffer;
	}


	//Returns container of containers of 
	std::vector<std::vector<Node<DIM>>> getNodesNeighbours()
	{
		std::vector<std::vector<Node<DIM>>> buffer; //buffer[i] is vector of neighbours of node with ID "i" 
		uint8_t fe_vertex_cnt = fin_els.begin()->nodes.size(); //we take size of finite element vector "nodes"(we take it from first element just for simplicity)
		uint8_t bfe_vertex_cnt = b_fin_els.begin()->second.nodes.size(); //we take size of boundary finite element vector "nodes"(we take it from first element just for simplicity)
		
		
		buffer.resize(nodes.size() + 1); //+1 because our indexation starts with 1
		
		for (auto it = fin_els.begin(); it != fin_els.end(); ++it)
		{
			for (uint8_t i = 0; i < fe_vertex_cnt; ++i)
			{
				for (uint8_t j = 0; j < fe_vertex_cnt; ++j)
				{
					if (i != j)
					{
						buffer[it->nodes[i]].push_back(getNodeWithID(it->nodes[j]));
					}
				}
			}
		}

		for (auto it = b_fin_els.begin(); it != b_fin_els.end(); ++it)
		{
			for (uint8_t i = 0; i < bfe_vertex_cnt; ++i)
			{
				for (uint8_t j = 0; j < bfe_vertex_cnt; ++j)
				{
					if (i != j)
					{
						buffer[it->second.nodes[i]].push_back(getNodeWithID(it->second.nodes[j]));
					}
				}
			}
		}

        //       ,
        //     
		auto comp = [](const Node<DIM>& n1, const Node<DIM>& n2)
		{
			if (n1.id < n2.id) return true;
			else return false;
		};

		auto equal = [](const Node<DIM>& n1, const Node<DIM>& n2)
		{
			if (n1.id == n2.id) return true;
			else return false;
		};

		for (auto& cur_vec : buffer)
		{
			std::sort(cur_vec.begin(), cur_vec.end(), comp);
			auto last = std::unique(cur_vec.begin(), cur_vec.end(), equal);
			cur_vec.erase(last, cur_vec.end()); //erase incorrect elements
		}

		return buffer;
	}


	//Returns Node with given ID from "nodes" container
	Node<DIM>& getNodeWithID(const uint32_t node_id)
	{

		std::vector<Node<DIM>>::template iterator node = std::find_if(nodes.begin(), nodes.end(),
                                                    // node_id    
													[&node_id](const Node<DIM>& cur_node)
													{
                                                        //    
														if (cur_node.id == node_id)
															return true;
														else
															return false;
													});

		if (node == nodes.end()) throw NodeWrongIDExc();
		else return *node;
	}


	void insertInCentres()
	{
		std::unordered_set<Edge, HashEdge> edges; //edges container

		for (auto& cur_fe : fin_els) //go through container of finite elements
		{
			std::vector<uint32_t> cur_fe_nodes = cur_fe.nodes; //current finite element nodes vector

			//Go through all edges of current finite element
			for (auto i = cur_fe_nodes.begin(); i != cur_fe_nodes.end() - 1; ++i)
			{
				for (auto j = i + 1; j != cur_fe_nodes.end(); ++j)
				{
					Edge cur_edge(*i, *j);

					auto cur_edge_it = edges.find(cur_edge); //find edge "cur_edge" in container of edges(this container contains edges that we have found before)

					if (cur_edge_it == edges.end()) //edge was not processed before
					{
						//Adding in nodes container
						Node<DIM> new_node = createMiddleNode(getNodeWithID(*i), getNodeWithID(*j));
						nodes.push_back(new_node);

						//Adding in edges container
						cur_edge.central_node_id = new_node.id;
						edges.insert(cur_edge);

						//Adding in current FE nodes vector
						cur_fe.nodes.push_back(cur_edge.central_node_id);
					}
					else
					{
						//Adding in current FE nodes vector
						cur_fe.nodes.push_back(cur_edge_it->central_node_id);
					}
				}
			}
		}

		for (auto& cur_bfe : b_fin_els) //go through container of boundary finite elements
		{
			std::vector<uint32_t> cur_bfe_nodes = cur_bfe.second.nodes;  //current boundary finite element nodes vector


			//Go through all edges of current finite element
			for (auto i = cur_bfe_nodes.begin(); i != cur_bfe_nodes.end() - 1; ++i)
			{
				for (auto j = i + 1; j != cur_bfe_nodes.end(); ++j)
				{
					Edge cur_edge(*i, *j);
					auto cur_edge_it = edges.find(cur_edge); //find edge "cur_edge" in container of edges(this container contains edges that we have found before)

					
					if (cur_edge_it == edges.end())
					{
						//Adding in nodes container
						Node<DIM> new_node = createMiddleNode(getNodeWithID(*i), getNodeWithID(*j));
						nodes.push_back(new_node);

						//Adding in edges container
						cur_edge.central_node_id = new_node.id;
						edges.insert(cur_edge);

						//Adding in current BFE nodes vector
						cur_bfe.second.nodes.push_back(cur_edge.central_node_id);
					}
					else
					{
						//Adding in current BFE nodes vector
						cur_bfe.second.nodes.push_back(cur_edge_it->central_node_id);
					}
				}
			}
		}
	}


	//Returns size of container "nodes"
	uint32_t getNodeCnt()
	{
		return nodes.size();
	}


	//Returns size of container "fin_els"
	uint32_t getFECnt()
	{
		return fin_els.size();
	}


	//Returns size of container "b_fin_els"
	uint32_t getBFECnt()
	{
		return b_fin_els.size();
	}
};
